%:-initialization(horpod).

horpod:-
	initialize,
	repeat,
	read(T),
	treatall(T),!,
	g_read(typeprec,Ltp),
        buildPrec(Ltp,FullTPrec),
	g_assign(typeprec,FullTPrec),
	g_read(funcprec,Lfp),
        buildPrec(Lfp,FullFPrec),
	g_assign(funcprec,FullFPrec),
	checkParams,
	g_read(rules,Lrules),
	reverse(Lrules,Rrules),
	g_assign(rules,Rrules),!,
	g_assign(crules,1),
	g_assign(termin,1),
	checkSN,!.

horpo:-
	initialize,
	write('Input file: '),
	read(Name),
	makefilename(Name,FName),
	see(FName),
	repeat,
	read(T),
	treatall(T),
	seen,!,
	g_read(typeprec,Ltp),
        buildPrec(Ltp,FullTPrec),
	g_assign(typeprec,FullTPrec),
	g_read(funcprec,Lfp),
        buildPrec(Lfp,FullFPrec),
	g_assign(funcprec,FullFPrec),
	checkParams,
	g_read(rules,Lrules),
	reverse(Lrules,Rrules),
	g_assign(rules,Rrules),!,
	g_assign(crules,1),
	g_assign(termin,1),
	checkSN,
        told,!.

initialize:-
        told,
	seen,
	op(600, xfy, ->),
	op(600, xfy, =),
	op(600, xfy, >),
	g_assign(noproof,0),
	g_assign(typecons,[]),
	g_assign(funcdecs,[]),
	g_assign(rules,[]),
	g_assign(typevars,[]),
	g_assign(typeprec,[]),
	g_assign(typelex,[[],[]]),
	g_assign(funcprec,[]),
	g_assign(funclex,[[],[]]).

makefilename(Name,Name):- 
    file_exists(Name).
makefilename(Name,FName):- 
    name(Name,Lname),
    name('.hor',Lext),
    append(Lname,Lext,Lfile),
    name(FName,Lfile),
    file_exists(FName).
makefilename(_,_):- 
    write('Error: File does not exist'),nl,nl,!,fail.

treatall(end_of_file).
treatall(T):-
    treat(T), fail.

treat(outfile(F)):-!,
        tell(F).
treat(noproof):-!,
	g_assign(noproof,1).
treat(typevars(LT)):-!,
        g_read(typevars,LT1),
        append(LT,LT1,LT2),
        sort(LT2,LT3),
	g_assign(typevars,LT3).
%treat(typecons(S,A)):-
%	nonvar(S),
%        atom(S),!,
%        op(600, fx, S),
%	g_read(typecons,L),
%	g_assign(typecons,[[S,A]|L]).
treat(typecons(S,A)):-!,
	nonvar(S),
	g_read(typecons,L),
	g_assign(typecons,[[S,A]|L]).
%treat(funcdec(F,IL,O)):-
%	nonvar(F),
%        atom(F),!,
%        op(600, fx, F),
%	g_read(funcdecs,L),
%	g_assign(funcdecs,[[F,IL,O]|L]).
treat(funcdec(F,IL,O)):-!,
	nonvar(F),
	g_read(funcdecs,L),
	g_assign(funcdecs,[[F,IL,O]|L]).
treat(rule(Env,Left,Right)):-!,
	g_read(rules,L),
	g_assign(rules,[[Env,Left,Right]|L]).
treat(typeprec(TP)):-!,
	g_read(typeprec,TP1),
        append(TP1,TP,TP2),
	g_assign(typeprec,TP2).
treat(typelex(LRLT,RLLT)):-!,
	g_read(typelex,[LRLT1,RLLT1]),
        append(LRLT,LRLT1,LRLT2),
        append(RLLT,RLLT1,RLLT2),
	g_assign(typelex,[LRLT2,RLLT2]).
treat(funcprec(FP)):-!,
	g_read(funcprec,FP1),
        append(FP1,FP,FP2),
	g_assign(funcprec,FP2).
treat(funclex(LRLF,RLLF)):-!,
	g_read(funclex,[LRLF1,RLLF1]),
        append(LRLF,LRLF1,LRLF2),
        append(RLLF,RLLF1,RLLF2),
	g_assign(funclex,[LRLF2,RLLF2]).
treat(T):-
        write('Input format error.'),nl,
        write(T),nl,!,fail.


buildPrec([],[]).
buildPrec([P1|LPrec],FullPrec):-
	buildPrec(LPrec,FullPrec1),
	add2Prec(P1,FullPrec1,FullPrec).

add2Prec(C>D,FullPrec1,FullPrec):-
	atomic(C),atomic(D),!,
	addgr(C,D,FullPrec1,FullPrec).
add2Prec(C=D,FullPrec1,FullPrec):-
	atomic(C),atomic(D),!,
	addeq(C,D,FullPrec1,FullPrec2),
	addeq(D,C,FullPrec2,FullPrec).
add2Prec(C>D,FullPrec1,FullPrec):-
	atomic(C),
    (D=..['=',E|_];D=..['>',E|_]),
	atomic(E),!,
	add2Prec(D,FullPrec1,FullPrec2),
	addgr(C,E,FullPrec2,FullPrec).
add2Prec(C=D,FullPrec1,FullPrec):-
	atomic(C),
	D=..[_,E|_],
	atomic(E),!,
	add2Prec(D,FullPrec1,FullPrec2),
	addeq(C,E,FullPrec2,FullPrec3),
	addeq(E,C,FullPrec3,FullPrec).

addgr(C,D,FullPrec1,FullPrec):-
    (member([D,Leq,Lgr],FullPrec1);[Leq,Lgr]=[[],[]]),
    append(Leq,Lgr,Ltot),
    addgr1(C,[D|Ltot],FullPrec1,FullPrec),!.

addgr1(C,Ltot,[],[[C,[],Ltot]]).
addgr1(C,Ltot,[[C,Leq,Lgr]|FullPrec1],[[C,Leq,Lgr2]|FullPrec]):-!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr2(C,Ltot,FullPrec1,FullPrec).
addgr1(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    (member(C,Leq);member(C,Lgr)),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr1(C,Ltot,FullPrec1,FullPrec).
addgr1(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addgr1(C,Ltot,FullPrec1,FullPrec).

addgr2(_,_,[],[]).
addgr2(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    (member(C,Leq);member(C,Lgr)),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr2(C,Ltot,FullPrec1,FullPrec).
addgr2(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addgr2(C,Ltot,FullPrec1,FullPrec).

addeq(C,D,FullPrec1,FullPrec):-
    (member([D,Leq,Lgr],FullPrec1);[Leq,Lgr]=[[],[]]),
    Leq1=[D|Leq],
    addgr1(C,Lgr,FullPrec1,FullPrec2),
    addeq1(C,Leq1,FullPrec2,FullPrec),!.


addeq1(_,_,[],[]).
addeq1(C,Ltot,[[C,Leq,Lgr]|FullPrec1],[[C,Leq2,Lgr]|FullPrec]):-!,
    append(Ltot,Leq,Leq1), sort(Leq1,Leq2),
    addeq2(C,Ltot,FullPrec1,FullPrec).
addeq1(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq2,Lgr]|FullPrec]):-
    member(C,Leq),!,
    append(Ltot,Leq,Leq1), sort(Leq1,Leq2),
    addeq1(C,Ltot,FullPrec1,FullPrec).
addeq1(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    member(C,Lgr),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addgr1(C,Ltot,FullPrec1,FullPrec).
addeq1(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addeq1(C,Ltot,FullPrec1,FullPrec).

addeq2(_,_,[],[]).
addeq2(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq2,Lgr]|FullPrec]):-
    member(C,Leq),!,
    append(Ltot,Leq,Leq1), sort(Leq1,Leq2),
    addeq2(C,Ltot,FullPrec1,FullPrec).
addeq2(C,Ltot,[[E,Leq,Lgr]|FullPrec1],[[E,Leq,Lgr2]|FullPrec]):-
    member(C,Lgr),!,
    append(Ltot,Lgr,Lgr1), sort(Lgr1,Lgr2),
    addeq2(C,Ltot,FullPrec1,FullPrec).
addeq2(C,Ltot,[S|FullPrec1],[S|FullPrec]):-
    addeq2(C,Ltot,FullPrec1,FullPrec).


checkParams:-
	g_read(typevars,L),
	member(X,L), var(X),
	write('Type variabels cannot start with capital letters'),nl,nl,!,fail.
checkParams:-
	g_read(typeprec,TP),
	checkPrecreflex(TP),
	write('The type precedence is not well-founded'),nl,nl,!,fail.
checkParams:-
	g_read(funcprec,FP),
	checkPrecreflex(FP),
	write('The function precedence is not well-founded'),nl,nl,!,fail.
checkParams:-
	g_read(typeprec,TP),
	g_read(typelex,[LRLT,RLLT]),
	\+(checkPrecStat(TP,LRLT,RLLT)),
	write('The type precedence is not compatible with the type statuses'),nl,nl,!,fail.
checkParams:-
	g_read(funcprec,FP),
	g_read(funclex,[LRLF,RLLF]),
	\+(checkPrecStat(FP,LRLF,RLLF)),
	write('The function precedence is not compatible with the function statuses'),nl,nl,!,fail.

checkParams.

checkPrecreflex([[C,_,Lgr]|_]):-
    member(C,Lgr),!.
checkPrecreflex([_|L]):-
    checkPrecreflex(L).

checkPrecStat([],_,_).
checkPrecStat([[G,Leq,_]|R],LRLT,RLLT):-
	checkPrecStat1(G,Leq,LRLT,RLLT),!,
	checkPrecStat(R,LRLT,RLLT).

checkPrecStat1(G,Leq,LRLT,_):-
    member(G,LRLT),!,
    subseteq(Leq,LRLT).
checkPrecStat1(G,Leq,_,RLLT):-
    member(G,RLLT),!,
    subseteq(Leq,RLLT).
checkPrecStat1(_,Leq,LRLT,RLLT):-
    append(LRLT,RLLT,LT),
    disjoint(Leq,LT).

subseteq([],_).
subseteq([X|L1],L2):-
    member(X,L2),
    subseteq(L1,L2).

disjoint([],_):-!.
disjoint([X|L1],L2):-
    \+(member(X,L2)),
    disjoint(L1,L2).

checkSN:-
    g_read(rules,[]),
    g_read(termin,0),
    write('Not all rules could be checked.'),nl,!.
checkSN:-
    g_read(rules,[]),
    g_read(termin,1),
    write('All rules checked. The system is SN'),nl,!.
checkSN:-
    g_read(rules,[[Env,Left,Right]|LR]),
    g_read(crules,N),
    write('Checking rule '), write(N), 
    \+(\+((numbervars([Env,Left,Right]),
	   g_assign(ruleenv,Env),
	   g_assign(ruleleft,Left),
	   g_assign(ruleright,Right)))),
    showrule,
    checkRule(Env,Left,Right),
    g_assign(rules,LR),
    N1 is N+1,
    g_assign(crules,N1),!,
    checkSN.

showrule:-
    g_read(noproof,0),    
    write(':'),nl,
    g_read(ruleenv,IEnv),
    g_read(ruleleft,ILeft), 
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(ILeft), write(' --> '), 
    write_lamterm(IRight),nl,nl.
showrule:-
    write(':'),nl,nl.


checkRule(Env,Left,Right):-
    write('  Checking left-hand side type... '),
    checkType(Env,Left,LeftT:LT),
    showresultok(LeftT:LT),
    checkRule1(Env,LeftT:LT,Right).
checkRule(_,_,_):-
    showresultfail, 
    g_assign(termin,0),
    write('  The left-hand side is not typable.'),nl,nl.

showresultok(LeftT:LT):- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleleft,ILeft),
    write(IEnv),write(' |-- '),
    write_lamterm(ILeft), nl,
    addtypes(ILeft,LeftT:LT,ILeftTW),
    write('  Resulting typed term:  '),
    write_lamterm(ILeftTW),nl,nl.
showresultok(_):- nl.

showresultfail:- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleleft,ILeft),
    write(IEnv),write(' |-- '),
    write_lamterm(ILeft), nl.
showresultfail:- nl.

checkRule1(Env,LeftT:LT,Right):-
    write('  Checking right-hand side type... '),
    checkType(Env,Right,RightT:RT),
    checkRule2(LeftT:LT,RightT:RT).
checkRule1(_,_,_):-
    showresult1fail,
    g_assign(termin,0),
    write('  The rigth-hand side is not typable.'),nl,nl.

showresult1fail:- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(IRight),nl.
showresult1fail:- nl.

checkRule2(LeftT:LT,RightT:RT):-
    subsumes([[RT,LT]],Sol),
    replacetype1term(RightT:RT,Sol,IRightT:IRT),
    showresult2ok(IRightT:IRT), 
    checkRule3(LeftT:LT,IRightT:IRT).
checkRule2(_,RightT:RT):-
    showresult2fail(RightT:RT), 
    g_assign(termin,0),
    write('  The types of the left-hand side and rigth-hand side are not compatible.'),nl,nl.

showresult2ok(IRightT:IRT):- 
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(IRight),nl,
    addtypes(IRight,IRightT:IRT,IRightW),
    write('  Resulting typed term:  '),
    write_lamterm(IRightW),nl,nl.
showresult2ok(_):- nl,nl.

showresult2fail(RightT:RT):-
    g_read(noproof,0),
    g_read(ruleenv,IEnv),
    g_read(ruleright,IRight),
    write(IEnv),write(' |-- '),
    write_lamterm(IRight),nl,
    addtypes(IRight,RightT:RT,IRightW),
    write('  Resulting typed term:  '),
    write_lamterm(IRightW),nl,nl.
showresult2fail(_):- nl.


checkRule3(LeftT:LT,IRightT:IRT):-
    write('  Checking ordering... '),nl,
    flush_output,!,
    checkruleord(LeftT:LT,IRightT:IRT,P),
    writeproof(P).

writeproof(fail):-
	g_assign(termin,0),
	write('The rule cannot be proved terminating.'),nl,nl.
writeproof(P):-
        g_read(noproof,0),
	write('   The rule is terminating.'),nl,
	write('   Proof:'),nl,
	prettyprint(3,P),nl,nl.
writeproof(_):-
	write('   The rule is terminating.'),nl,nl.

checkType(Env,X,X:T):-
    var(X),!,
    memberenv(X:T,Env).
checkType(Env,X,X:T):-
    atomic(X),
    memberenv(X:T,Env).
%checkType(_,X,X:T):-
%    atomic(X),!,
%    g_read(funcdecs,LFD),
%    member([X,[],T],LFD).
checkType(Env,lam(X:T1,U),lam(X1,V1:T2):(T1->T2)):-
    memberenv(X:_,Env),!,
    (nonvar(X)-> (name(X,L),append(L,[49],L1),name(X1,L1));true),
    replacebvar(X,X1,U,U1),
    checkType([X1:T1|Env],U1,V1:T2).
checkType(Env,lam(X:T1,U),lam(X,V:T2):(T1->T2)):-!,
    checkType([X:T1|Env],U,V:T2).
checkType(Env,lam(X,U),lam(X1,V1:T2):(T1->T2)):-
    memberenv(X:_,Env),!,
    (nonvar(X)-> (name(X,L),append(L,[49],L1),name(X1,L1));true),
    replacebvar(X,X1,U,U1),
    checkType([X1:T1|Env],U1,V1:T2).
checkType(Env,lam(X,U),lam(X,V:T2):(T1->T2)):-!,
    checkType([X:T1|Env],U,V:T2).
checkType(Env,Term,Term1:T):-
    Term=..[@,U|Args],!,
    checkType(Env,U,U1:T1),
    checkTypelist(Env,Args,Args1),
    length(Args,N),
    checkInstArrow(N,T1,Args1,IT1,IArgs1),
    getOutType(N,IT1,T),
    Term1=..[@,U1:T1|IArgs1].
checkType(Env,Term,Term1:T):-!,
    Term=..[F|Args],
    length(Args,NA),
    g_read(funcdecs,LFD),
    member([F,IL,O],LFD),
    length(IL,NA),
    checkTypelist(Env,Args,Args1),
    insttypelist(IL,Args1,O,IArgs1,T),
    Term1=..[F|IArgs1].
checkType(_,_,_):-
	write('The term can not be typed. Fail.'),nl,nl,!,fail.

%checkType([x:real,F:(real->real)],@(der(F),x),real,T).

checkTypelist(_Env,[],[]).
checkTypelist(Env,[A1|Args],[A2|Args1]):-
    checkType(Env,A1,A2),
    checkTypelist(Env,Args,Args1).

memberenv(X:T1,[Y:T2|_]):-
    X==Y,!,
    eqType(T1,T2),!.
memberenv(X:T,[_|Env]):-
    memberenv(X:T,Env).

membervar(X,[Y|_]):-
    X==Y,!.
membervar(X,[_|L]):-
    membervar(X,L).

getOutType(0,T,T):-!.
getOutType(N,T,O):-
    N1 is N -1,
    T=..[->,_,T1],!,
    getOutType(N1,T1,O).

getInTypes(0,_,[]):-!.
getInTypes(N,T,[I|R]):-
    N1 is N -1,
    T=..[->,I,T1],!,
    getInTypes(N1,T1,R).

getTypeslist([],[]).
getTypeslist([_:T1|Args],[T1|IL1]):-
	getTypeslist(Args,IL1).

replacebvar(X,X1,Y,X1):-X==Y,!.
replacebvar(_,_,Y,Y):-var(Y).
replacebvar(X,_,lam(Y,U),lam(Y,U)):-X==Y,!.
replacebvar(X,X1,lam(Y,U),lam(Y,U1)):-!,
    replacebvar(X,X1,U,U1).
replacebvar(X,X1,U,U1):-
    U=..[F|Args],
    replacebvarlist(X,X1,Args,Args1),
    U1=..[F|Args1].  

replacebvarlist(_,_,[],[]).
replacebvarlist(X,X1,[U|Args],[U1|Args1]):-
    replacebvar(X,X1,U,U1),
    replacebvarlist(X,X1,Args,Args1).

checkInstArrow(N,T,Args,IT,IArgs):-
    g_read(typevars,LTV),
    g_read(typecons,LTC),
    getTypeslist(Args,IAL),
    getInTypes(N,T,TL),
    makeeqs(TL,IAL,Prob),
    mgu(Prob,LTV,LTC,Sol),
    replacetype1termlist(Args,Sol,IArgs),
    replacetype1(T,Sol,IT).

insttypelist(IL,Args,O,IArgs,T):-
    g_read(typevars,LTV),
    g_read(typecons,LTC),
    getTypeslist(Args,IAL),
    makeeqs(IL,IAL,Prob),
    mgu(Prob,LTV,LTC,Sol),
    replacetype1termlist(Args,Sol,IArgs),
    replacetype1(O,Sol,T).

makeeqs([],[],[]).
makeeqs([T1|L1],[T2|L2],[T11=T12|Prob]):-
    sort0([T1,T2],[T11,T12]),
    makeeqs(L1,L2,Prob).

mgu([],_,_,[]).
mgu(Prob,LTV,LTC,Sol):-
    delete1(T=T,Prob,Prob1),!,
    mgu(Prob1,LTV,LTC,Sol),!.
mgu(Prob,_,LTC,_):-
    delete1(T1=T2,Prob,_),
    T1=..[C1|_],
    membereqsol([C1,_],[[->,_]|LTC]),
    T2=..[C2|_],
    membereqsol([C2,_],[[->,_]|LTC]),
    \+(C1=C2),!,fail.
mgu(Prob,LTV,LTC,Sol):-
    delete1((T11->T12)=(T21->T22),Prob,Prob1),!,
    mgu([T11=T21,T12=T22|Prob1],LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,Sol):-
    delete1(T1=T2,Prob,Prob1),
    T1=..[C|LT1],
    membereqsol([C,_],LTC),
    T2=..[C|LT2],!,
    makeeqs(LT1,LT2,Prob2),
    append(Prob2,Prob1,Prob3),
    mgu(Prob3,LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,[[T2,T1]|Sol]):-
    delete1(T1=T2,Prob,Prob1),
    membereq(T1,LTV),
    membereq(T2,LTV),!,
    replacetype(Prob1,[T2,T1],Prob2),
    mgu(Prob2,LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,Sol):-
    delete1(T1=T2,Prob,Prob1),
    membereq(T1,LTV),
    \+(membereq(T2,LTV)),
    (delete(T1=T3,Prob1,Prob2);
     delete(T3=T1,Prob1,Prob2)),
    sort0([T2,T3],[A1,A2]),
    mgu([T1=A1,A1=A2|Prob2],LTV,LTC,Sol),!.
mgu(Prob,LTV,LTC,Sol):-
    delete1(T2=T1,Prob,Prob1),
    membereq(T1,LTV),
    \+(membereq(T2,LTV)),
    (delete(T1=T3,Prob1,Prob2);
     delete(T3=T1,Prob1,Prob2)),
    sort0([T2,T3],[A1,A2]),
    mgu([T1=A1,A1=A2|Prob2],LTV,LTC,Sol),!.
mgu(Prob,LTV,_,Sol):-
    mgusol(Prob,LTV,Sol),!.

mgusol([],_,[]).
mgusol([T2=T1|Prob],LTV,[[T1,T2]|Sol1]):-
    membereq(T1,LTV),!,
    mgusol(Prob,LTV,Sol),!,
    replacetype1(T2,Sol,T3),!,
    \+(subt(T1,T3)),
    replaceinsol(T1,T2,Sol,Sol1).
mgusol([T1=T2|Prob],LTV,[[T1,T2]|Sol1]):-
    mgusol(Prob,LTV,Sol),!,
    replacetype1(T2,Sol,T3),!,
    \+(subt(T1,T3)),
    replaceinsol(T1,T2,Sol,Sol1).


subt(X,X):-!.
subt(X,T):-
    T=..[_|Args],!,
    subtlist(X,Args),!.

subtlist(X,[T|_]):-
    subt(X,T),!.
subtlist(X,[_|Args]):-
    subtlist(X,Args),!.

subsumes(Prob,Sol):-
    g_read(typevars,LTV),
    g_read(typecons,LTC),
    match(Prob,LTV,LTC,Sol1),
    sort(Sol1,Sol),
    matchsol(Sol).

match([],_,_,[]).
match([[T,T]|Prob],LTV,LTC,Sol):-
    match(Prob,LTV,LTC,Sol).
match([[(T11->T12),(T21->T22)]|Prob],LTV,LTC,Sol):-
    match([[T11,T21],[T12,T22]|Prob],LTV,LTC,Sol).
match([[T1,T2]|Prob],LTV,LTC,Sol):-
    T1=..[C|LT1],
    membereqsol([C,_],LTC),
    T2=..[C|LT2],!,
    makeeqs(LT1,LT2,Prob1),
    append(Prob1,Prob,Prob2),
    match(Prob2,LTV,LTC,Sol).
match([[T1,T2]|Prob],LTV,LTC,[[T1,T2]|Sol]):-
    membereq(T1,LTV),!,
    match(Prob,LTV,LTC,Sol).

matchsol([]).
matchsol([_]).
matchsol([[X,_],[Y,T2]|L]):-
    X\=Y,!,
    matchsol([[Y,T2]|L]).

replacetype1term(T:I,Sol,T:I1):-
	   var(T),!,
	   replacetype1(I,Sol,I1).
replacetype1term(T:I,Sol,T1:I1):-
	   replacetype1(I,Sol,I1),
	   T=..[O|Args],
	   replacetype1termlist(Args,Sol,Args1),
	   T1=..[O|Args1].

replacetype1termlist([],_,[]).
replacetype1termlist([T:I|RT],Sol,[T1:I1|RT1]):-
    replacetype1term(T:I,Sol,T1:I1),
    replacetype1termlist(RT,Sol,RT1).
replacetype1termlist([X|RT],Sol,[X|RT1]):-
    atom(X),
    replacetype1termlist(RT,Sol,RT1).

replacetype([],_,[]).
replacetype([I|IL],Sol,[I1|IL1]):-
    replacetype1(I,Sol,I1),
    replacetype(IL,Sol,IL1).

replaceinsol(_,_,[],[]).
replaceinsol(T1,T2,[[V,I]|IL],[[V,I1]|IL1]):-
    replacetype1(I,[T1,T2],I1),
    replaceinsol(T1,T2,IL,IL1).

replacetype1(I,Sol,I1):-
    membereqsol([I,I1],Sol),!.
replacetype1(I,_,I):-
    var(I),!.
replacetype1(I,_,I):-
    atomic(I),!.
replacetype1(I,Sol,I1):-
    I=..[C|AI],
    replacetype(AI,Sol,AI1),
    I1=..[C|AI1].

membereq(X,[Y|_]):-
    X==Y,!.
membereq(X,[_|L]):-
    membereq(X,L).

membereq2(X,[[Y|_]|_]):-
    X==Y,!.
membereq2(X,[_|L]):-
    membereq2(X,L).

membereqsol([X,Z],[[Y,Z]|_]):-
    X==Y,!.
membereqsol(P,[_|L]):-
    membereqsol(P,L).

addtypes('$VAR'(N),_:T,'$VAR'(N):T):-!.
addtypes(lam(X,U),lam(_,V):T,lam(X,W):T):-!,
	addtypes(U,V,W).
addtypes(IT,TT:T,ITT:T):-
	IT=..[F|ArgsI],
	TT=..[F|ArgsT],
	addtypeslist(ArgsI,ArgsT,ArgsIT),
	ITT=..[F|ArgsIT].

addtypeslist([],[],[]).
addtypeslist([X|R1],[Y|R2],[Z|R3]):-
	addtypes(X,Y,Z),
	addtypeslist(R1,R2,R3).


checkruleord(Left:T1,Right:T2,Proof):-
	checkhorpo(Left:T1,Right:T2,Proof).
checkruleord(_,_,fail).


checkhorpoge(A1,A2,equal):-
    eqlamTerm(A1,A2).
checkhorpoge(A1,A2,Proof):-
    checkhorpo(A1,A2,Proof).

checkhorpo(Left:T1,Right:T2,Proof):-
    nonvar(Left),
    geType(T1,T2),!,
    checkhorpo1(Left,T1,Right,T2,Proof).

checkhorpo1(Left,T1,Right,T2,P):-
    Left=..[F|Args],
    g_read(funcdecs,LFD),
    membereq2(F,LFD),!,
    checkhorpo1Func(F,Args,T1,Right,T2,P).
checkhorpo1(Left,T1,Right,T2,P):-
    Left=..[@|_],!,
    flatApp(Left,Args),
    checkhorpo1App(Args,T1,Right,T2,P).
checkhorpo1(lam(X,U),T1,Right,T2,P):-!,
    checkhorpo1Lam(lam(X,U),T1,Right,T2,P).

checkhorpo1Func(_F,Args,_,Right,T2,p(1,N,equal)):- %Case 1 Subterm
    nth(N,Args,A),
    eqlamTerm(A,Right:T2),!.
checkhorpo1Func(F,Args,T1,Right,_,p(7,LP)):- %Case 7 FuncApp
    nonvar(Right),
    Right=..[@|Args1],
    checkhorpo1FuncList(F,Args,T1,Args1,LP),!.
checkhorpo1Func(F,Args,T1,Right,_,p(8,P)):- %Case 8 FuncLam
    nonvar(Right),
    Right=lam(Y,V:T3), 
    notin(Y,V),
    checkhorpo1Func(F,Args,T1,V,T3,P),!.
checkhorpo1Func(F,Args,T1,Right,_,p(2,LP)):- %Case 2 Prec 
    nonvar(Right),
    Right=..[F1|Args1],F1 \== @, F1 \== lam,
    g_read(funcdecs,LFD),
    membereq2(F1,LFD),
    gtPrecF(F,F1),
    checkhorpo1FuncList(F,Args,T1,Args1,LP),!.
checkhorpo1Func(F,Args,T1,Right,_,p(S,LP)):- %Case 3 Mul %Case 4 Lex
    nonvar(Right),
    Right=..[F1|Args1],F1 \== @, F1 \== lam,
    g_read(funcdecs,LFD),
    membereq2(F1,LFD),
    eqPrecT(F,F1),
    checkhorpo1FuncStat(F,Args,T1,Args1,S,LP),!.
checkhorpo1Func(_,Args,_,Right,T2,p(1,N,P)):- %Case 1 Subterm
    nth(N,Args,A),
    checkhorpo(A,Right:T2,P),!.
checkhorpo1Func(F,Args,T1,Right,T2,p(1,cc,P)):- %Case 1 Subterm
    geCC(F,Args,T1,Right,T2,P),!.


checkhorpo1FuncStat(C,Args1,T,Args2,4-lr,LP):-
    g_read(funclex,[L1,_]),
    member(C,L1),
    checkhorpoLex(1,Args1,Args2,Args3,N3,LP1),
    (Args3=[] ->
	LP=LP1 ;
	(checkhorpo1FuncList(C,Args1,T,Args3,LP2),
	    append(LP1,[restargs,N3,LP2],LP))).
checkhorpo1FuncStat(C,Args1,T,Args2,4-rl,LP):-
    g_read(funclex,[_,L2]),
    member(C,L2),
    reverse(Args1,IArgs1),
    reverse(Args2,IArgs2),
    checkhorpoLex(0,IArgs1,IArgs2,Args3,N3,LP1),
    (Args3=[] ->
	LP=LP1 ;
	(checkhorpo1FuncList(C,Args1,T,Args3,LP2),
	    append(LP1,[restargs,N3,LP2],LP))).
checkhorpo1FuncStat(_,Args1,_,Args2,3,LP):-
    checkhorpoMul(1,Args1,Args2,LP).


checkhorpoLex(N,[A1|_],[A2|Args2],Args2,N3,[p(N,P)]):-
	     N3 is N+1,
	     checkhorpo(A1,A2,P),!.
checkhorpoLex(N,[A1|Args1],[A2|Args2],Args3,N3,[p(N,equal)|LP]):-
	     N1 is N+1,
	     eqlamTerm(A1,A2),!,
	     checkhorpoLex(N1,Args1,Args2,Args3,N3,LP),!.

checkhorpoMul(_,L,[],[]):-member(X,L),X\=[],!.
checkhorpoMul(N,Args1,[A2|Args2],[p(N1,N,equal)|LP]):-
	nthdelete1(N1,A1,Args1,Args3),
	eqlamTerm(A1,A2),!,
	N2 is N+1,
	checkhorpoMul(N2,Args3,Args2,LP).
checkhorpoMul(N,Args1,[A2|Args2],[p(N1,N,P)|LP]):-
	nth(N1,Args1,A1),
	checkhorpo(A1,A2,P),!,
	N2 is N+1,
	checkhorpoeqMul(N2,Args1,Args2,LP).

checkhorpoeqMul(_,_,[],[]).
checkhorpoeqMul(N,Args1,[A2|Args2],[p(N1,N,equal)|LP]):-
	nthdelete1(N1,A1,Args1,Args3),
	eqlamTerm(A1,A2),!,
	N2 is N+1,
	checkhorpoeqMul(N2,Args3,Args2,LP).
checkhorpoeqMul(N,Args1,[A2|Args2],[p(N1,N,P)|LP]):-
	nth(N1,Args1,A1),
	checkhorpo(A1,A2,P),!,
	N2 is N+1,
	checkhorpoeqMul(N2,Args1,Args2,LP).

checkhorpo1FuncList(_,_,_,[],[]). 
checkhorpo1FuncList(C,Args1,T1,[A2:T2|Args2],[p(0,P)|LP]):-
	geType(T1,T2),%!,
	checkhorpo1Func(C,Args1,T1,A2,T2,P),
	checkhorpo1FuncList(C,Args1,T1,Args2,LP).
checkhorpo1FuncList(C,Args1,T1,[A2:T2|Args2],[p(N,equal)|LP]):-
	nth(N,Args1,A1),
        eqlamTerm(A1,A2:T2),
        checkhorpo1FuncList(C,Args1,T1,Args2,LP),!.
checkhorpo1FuncList(C,Args1,T1,[A2:T2|Args2],[p(N,P)|LP]):-
	nth(N,Args1,A1),
        checkhorpo(A1,A2:T2,P),
        checkhorpo1FuncList(C,Args1,T1,Args2,LP),!.
checkhorpo1FuncList(C,Args1,T1,[A2:T2|Args2],[p(cc,P)|LP]):-
        geCC(C,Args1,T1,A2,T2,P),!,
        checkhorpo1FuncList(C,Args1,T1,Args2,LP).

notin(X,Y):-
	(var(Y);atomic(Y)),!,
	X \== Y,!.
notin(X,lam(Y,_)):-
	X == Y,!.
notin(X,lam(_,U:_)):-!,
	notin(X,U),!.
notin(X,T):-!,
	T=..[_|Args],!,
	notinlist(X,Args),!.

notinlist(_,[]).
notinlist(X,[A:_|Args]):-
	notin(X,A),!,
	notinlist(X,Args).

checkhorpo1App(Args,_,Right,T2,p(5,N,equal)):- %Case 5 SubtApp
	nth(N,Args,A),
	eqlamTerm(A,Right:T2),!.
checkhorpo1App(Args,T1,Right,T2,p(9,pf(LN),LP)):- %Case 9 MulApp
	nonvar(Right),
	Right=..[@|Args2],
	makeapp2(Args,T1,@(A1,A2)),
	getpartflat(Args2,T2,PF),
	length(PF,LN),
	checkhorpoMul(1,[A1,A2],PF,LP),!.
checkhorpo1App(Args,_,Right,T2,p(5,N,P)):- %Case 5 SubtApp
	nth(N,Args,A),
	checkhorpo(A,Right:T2,P),!.
checkhorpo1App([A1,A2],T1,Right,T2,p(11,P)):- %Case 11 beta
	nonvar(A1),
	A1=lam(X,U),!,
	replace(X,A2,U,W),
	checkhorpo(W:T1,Right:T2,P),!.
checkhorpo1App([A1:(_->T),A2:_|Args],T1,Right,T2,p(11,P)):- %Case 11 beta
	nonvar(A1),
	A1=lam(X,U),!,
	replace(X,A2,U,W),
	checkhorpo1App([W:T|Args],T1,Right,T2,P),!.
checkhorpo1App(Args,T1,Right,T2,p(13,p(9,pf(2),LP))):- %Case 13 interm Beta
        nonvar(Right),
        Right=..[F|Args1],
        g_read(funcdecs,LFD),
        membereq2(F,LFD),
        makeapp2(Args,T1,@(A1,A2)),
        append(Args11,[Ult:Tu],Args1),
        append(Args11,[X:Tu],AArgs1),
        ARight=..[F|AArgs1],
        checkhorpoMul(1,[A1,A2],[lam(X,ARight:T2):(Tu->T2),Ult:Tu],LP),!.
        
makeapp2(Args,T1,@(A1:(Tf->T1),Af:Tf)):-
	append(Args1,[Af:Tf],Args),
	makeapp(Args1,A1).
makeapp([A1:_],A1):-!.
makeapp(Args1,A1):-
	A1=..[@|Args1],!.

getpartflat(Args,T2,[Ap1:T,A2|R2]):-
	append([A1|R1],[A2|R2],Args),
	extracttype([A2|R2],T2,T),
	makeapp([A1|R1],Ap1).

extracttype([_:T1],T2,(T1->T2)):-!.
extracttype([_:T1|R2],T2,(T1->T)):-
	extracttype(R2,T2,T).

replace(X,T,Y:O,T:O):-
	X==Y,!.
replace(_,_,U:O,U:O):-
	(var(U);atomic(U)),!.
replace(X,_,lam(Y,V):O,lam(Y,V):O):-
	X==Y,!.
replace(X,T,lam(Y,V):O,lam(Y,W):O):-
	replace(X,T,V,W),!.
replace(X,T,U:O,W:O):-
	U=..[F|Args],
	replacelist(X,T,Args,Args1),
	W=..[F|Args1],!.

replacelist(_,_,[],[]).
replacelist(X,T,[A|Args],[A1|Args1]):-
	replace(X,T,A,A1),
	replacelist(X,T,Args,Args1).

checkhorpo1Lam(lam(X,U),_,Right,T2,p(6,equal)):-  %Case  6 SubtLam
	replace(X,_,U,V),
	eqlamTerm(V,Right:T2),!.
checkhorpo1Lam(lam(X,U),_,Right,T2,p(6,P)):-  %Case  6 SubtLam
	replace(X,_,U,V),
	checkhorpo(V,Right:T2,P).
checkhorpo1Lam(lam(X,U),(TI1->_),Right,T2,p(10,P)):-  %Case 10 MonLam
	nonvar(Right),
	nonvar(T2),
	Right=lam(Y,V),
	T2=(TI2->_),
	eqType(TI1,TI2),!,
	replace(Y,X,V,W),
	checkhorpo(U,W,P).
checkhorpo1Lam(lam(X,U),T1,Right,T2,p(12,P)):-  %Case 12 Eta
	nonvar(U),
	U=..[@|Args],
	append(Args,Args1,[Last]),
	Last==U,
	notinlist(X,Args1),!,
	makeapp(Args1,A),
	checkhorpo1(A,T1,Right,T2,P),!.
checkhorpo1Lam(lam(X,U),(TI1->_),Right,T2,p(14,P)):-  %Case 14 interm Eta
	nonvar(T2),
	T2=(TI2->TO2),!,
	eqType(TI1,TI2),!,
	replace(X,NV,U,V),
	addapp(Right,T2,NV:TI2,A),
	checkhorpo(V,A:TO2,P).

addapp(A1,T,A2,@(A1:T,A2)):-
	var(A1),!.
addapp(A1,_,A2,A):-
	A1=..[@|Args1],!,
	append(Args1,[A2],Args),
	A=..[@|Args].
addapp(A1,T,A2,@(A1:T,A2)):-!.


eqCC(F,Args,T,Right,Tr,P):-
    eqCC(F,Args,T,Right,Tr,[],P).

eqCC(_,_,_,Right,_,BVars,pcc(bvar,base)):-
    membervar(Right,BVars),!.
eqCC(_,Args,_,Right,_,_,pcc(N,base)):-
    nth(N,Args,T:_),T==Right.
eqCC(F,Args,T1,Right,_,BVars,pcc(abs,ren(X,Y),P)):- %abstraction
    nonvar(Right),
    Right=lam(X,U),
    replace(X,Y,U,V:Tv),
    eqCC(F,Args,T1,V,Tv,[Y|BVars],P),!.
eqCC(F,Args,T1,Right,_,BVars,pcc(app,LP)):- %application
    nonvar(Right),
    Right=..[@|Args1],
    eqCClist(F,Args,T1,Args1,BVars,LP),!.
eqCC(F,Args,T1,Right,_,BVars,pcc(prec,LP)):- %precedence
    nonvar(Right),
    Right=..[G|Args1],G \= @, G \= lam,
    gtPrecF(F,G),
    eqCClist(F,Args,T1,Args1,BVars,LP),!.
eqCC(F,Args,T1,Right,_,BVars,pcc(reccall,S,OP,LP)):- %recursive-call
    nonvar(Right),
    Right=..[G|Args1],G \= @,G \= lam,
    eqPrecF(F,G),
    checkhorposubt1FuncStat(F,Args,Args1,S,OP),
    eqCClist(F,Args,T1,Args1,BVars,LP),!.
eqCC(_,Args,_,Right,Tr,_,pcc(red-base,N,OP)):- %reduction
    nth(N,Args,T),
    checkhorponcc(T,Right:Tr,OP),!.

eqCClist(_,_,_,[],_,[]).
eqCClist(F,Args,T,[A1:T1|Args1],BVars,[P|LP]):-
	eqCC(F,Args,T,A1,T1,BVars,P),
	eqCClist(F,Args,T1,Args1,BVars,LP).

geCC(F,Args,T,Right,Tr,p(cc,equal,P)):-
    eqCC(F,Args,T,Right,Tr,P).

%geCC(F,Args,T,Right,Tr,p(cc,Right1:T1,P1,P)):-
%    eqCC(F,Args,T,Right,Tr,Right1,T1,P1),
%    checkhorpoge(Right1:T1,Right:Tr,P).


%%%%%%%%%%%%%%%%%%%%%%%HORPO union substgetype%%%%%%%%%%%%%%%%%%%%%%%%%%%



checkhorposubt1FuncStat(C,Args1,Args2,lex-lr,LP):-
    g_read(funclex,[L1,_]),
    member(C,L1),
    checkhorposubtLex(1,Args1,Args2,LP).
checkhorposubt1FuncStat(C,Args1,Args2,lex-rl,LP):-
    g_read(funclex,[_,L2]),
    member(C,L2),
    reverse(Args1,IArgs1),
    reverse(Args2,IArgs2),
    checkhorposubtLex(0,IArgs1,IArgs2,LP).
checkhorposubt1FuncStat(_,Args1,Args2,mul,LP):-
    checkhorposubtMul(1,Args1,Args2,LP).


checkhorposubtLex(N,[A1|_],[A2|_],[p(N,P)]):-
	     checkhorponcc(A1,A2,P),!.
checkhorposubtLex(N,[A1|_],[A2|_],[p(N,substgT)]):-
	     subtgetype(A1,A2),!.
checkhorposubtLex(N,[A1|Args1],[A2|Args2],[p(N,equal)|LP]):-
	     eqlamTerm(A1,A2),!,
             N1 is N + 1,
	     checkhorposubtLex(N1,Args1,Args2,LP),!.

checkhorposubtMul(_,L,[],[]):-member(X,L),X\=[],!.
checkhorposubtMul(N,Args1,[A2|Args2],[p(N1,N,equal)|LP]):-
	nthdelete1(N1,A1,Args1,Args3),
	eqlamTerm(A1,A2),!,
	N2 is N+1,
	checkhorposubtMul(N2,Args3,Args2,LP).
checkhorposubtMul(N,Args1,[A2|Args2],[p(N1,N,P)|LP]):-
	nth(N1,Args1,A1),
	checkhorponcc(A1,A2,P),!,
	N2 is N+1,
	checkhorposubteqMul(N2,Args1,Args2,LP).
checkhorposubtMul(N,Args1,[A2|Args2],[p(N1,N,substgT)|LP]):-
	nth(N1,Args1,A1),
	subtgetype(A1,A2),!,
	N2 is N+1,
	checkhorposubteqMul(N2,Args1,Args2,LP).

checkhorposubteqMul(_,_,[],[]).
checkhorposubteqMul(N,Args1,[A2|Args2],[p(N1,N,equal)|LP]):-
	nthdelete1(N1,A1,Args1,Args3),
	eqlamTerm(A1,A2),!,
	N2 is N+1,
	checkhorposubteqMul(N2,Args3,Args2,LP).
checkhorposubteqMul(N,Args1,[A2|Args2],[p(N1,N,P)|LP]):-
	nth(N1,Args1,A1),
	checkhorponcc(A1,A2,P),!,
	N2 is N+1,
	checkhorposubteqMul(N2,Args1,Args2,LP).
checkhorposubteqMul(N,Args1,[A2|Args2],[p(N1,N,substgT)|LP]):-
	nth(N1,Args1,A1),
	subtgetype(A1,A2),!,
	N2 is N+1,
	checkhorposubteqMul(N2,Args1,Args2,LP).

subtgetype(A:_,_):-
    var(A),!,fail.
subtgetype(A:T1,B:T2):-
    A=..[@|Args],!,
    makeapp2(Args,T1,@(A1,A2)),
    (subteq(A1,B:T2);subteq(A2,B:T2)),!,
    geType(T1,T2),!.
subtgetype(lam(X,U):T1,B:T2):-!,
    subteq(U,B:T2),!,
    notin(X,B),!,
    geType(T1,T2),!.
subtgetype(A:T1,B:T2):-
    A=..[_|Args],
    subteqList(Args,B:T2),
    geType(T1,T2).

subteq(A:_,B:_):-
    var(A),!,A==B.
subteq(A:_,B:_):-
    A==B.
subteq(A:T1,B):-
    A=..[@|Args],
    makeapp2(Args,T1,@(A1,A2)),
    (subteq(A1,B);subteq(A2,B)).
subteq(lam(X,U):_,B:T2):-
    subteq(U,B:T2),
    notin(X,B).
subteq(A:_,B):-
    A=..[_|Args],
    subteqList(Args,B).

subteqList([A|_],B):-
    subteq(A,B).
subteqList([_|Args],B):-
    subteqList(Args,B).


%%%%%%%%%%%%%%%%%%%%%%%HORPO without CC%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

checkhorponcc(Left:T1,Right:T2,Proof):-
    nonvar(Left),
    geType(T1,T2),!,
    checkhorponcc1(Left,T1,Right,T2,Proof).

checkhorponcc1(Left,T1,Right,T2,P):-
    Left=..[F|Args],
    g_read(funcdecs,LFD),
    membereq2(F,LFD),!,
    checkhorponcc1Func(F,Args,T1,Right,T2,P).
checkhorponcc1(Left,T1,Right,T2,P):-
    Left=..[@|_],!,
    flatApp(Left,Args),
    checkhorponcc1App(Args,T1,Right,T2,P).
checkhorponcc1(lam(X,U),T1,Right,T2,P):-!,
    checkhorponcc1Lam(lam(X,U),T1,Right,T2,P).

checkhorponcc1Func(_F,Args,_,Right,T2,p(1,N,equal)):- %Case 1 Subterm
    nth(N,Args,A),
    eqlamTerm(A,Right:T2),!.
checkhorponcc1Func(F,Args,T1,Right,_,p(7,LP)):- %Case 7 FuncApp
    nonvar(Right),
    Right=..[@|Args1],
    checkhorponcc1FuncList(F,Args,T1,Args1,LP),!.
checkhorponcc1Func(F,Args,T1,Right,_,p(8,P)):- %Case 8 FuncLam
    nonvar(Right),
    Right=lam(Y,V:T3), 
    notin(Y,V),
    checkhorponcc1Func(F,Args,T1,V,T3,P),!.
checkhorponcc1Func(F,Args,T1,Right,_,p(2,LP)):- %Case 2 Prec 
    nonvar(Right),
    Right=..[F1|Args1],F1 \== @, F1 \== lam,
    g_read(funcdecs,LFD),
    membereq2(F1,LFD),
    gtPrecF(F,F1),
    checkhorponcc1FuncList(F,Args,T1,Args1,LP),!.
checkhorponcc1Func(F,Args,T1,Right,_,p(S,LP)):- %Case 3 Mul %Case 4 Lex
    nonvar(Right),
    Right=..[F1|Args1],F1 \== @, F1 \== lam,
    g_read(funcdecs,LFD),
    membereq2(F1,LFD),
    eqPrecT(F,F1),
    checkhorponcc1FuncStat(F,Args,T1,Args1,S,LP),!.
checkhorponcc1Func(_,Args,_,Right,T2,p(1,N,P)):- %Case 1 Subterm
    nth(N,Args,A),
    checkhorponcc(A,Right:T2,P),!.


checkhorponcc1FuncStat(C,Args1,T,Args2,4-lr,LP):-
    g_read(funclex,[L1,_]),
    member(C,L1),
    checkhorponccLex(1,Args1,Args2,Args3,N3,LP1),
    (Args3=[] ->
	LP=LP1 ;
	(checkhorponcc1FuncList(C,Args1,T,Args3,LP2),
	    append(LP1,[restargs,N3,LP2],LP))).
checkhorponcc1FuncStat(C,Args1,T,Args2,4-rl,LP):-
    g_read(funclex,[_,L2]),
    member(C,L2),
    reverse(Args1,IArgs1),
    reverse(Args2,IArgs2),
    checkhorponccLex(0,IArgs1,IArgs2,Args3,N3,LP1),
    (Args3=[] ->
	LP=LP1 ;
	(checkhorponcc1FuncList(C,Args1,T,Args3,LP2),
	    append(LP1,[restargs,N3,LP2],LP))).
checkhorponcc1FuncStat(_,Args1,_,Args2,3,LP):-
    checkhorponccMul(1,Args1,Args2,LP).


checkhorponccLex(N,[A1|_],[A2|Args2],Args2,N3,[p(N,P)]):-
	     N3 is N+1,
	     checkhorponcc(A1,A2,P),!.
checkhorponccLex(N,[A1|Args1],[A2|Args2],Args3,N3,[p(N,equal)|LP]):-
	     N1 is N+1,
	     eqlamTerm(A1,A2),!,
	     checkhorponccLex(N1,Args1,Args2,Args3,N3,LP),!.

checkhorponccMul(_,L,[],[]):-member(X,L),X\=[],!.
checkhorponccMul(N,Args1,[A2|Args2],[p(N1,N,equal)|LP]):-
	nthdelete1(N1,A1,Args1,Args3),
	eqlamTerm(A1,A2),!,
	N2 is N+1,
	checkhorponccMul(N2,Args3,Args2,LP).
checkhorponccMul(N,Args1,[A2|Args2],[p(N1,N,P)|LP]):-
	nth(N1,Args1,A1),
	checkhorponcc(A1,A2,P),!,
	N2 is N+1,
	checkhorpoeqnccMul(N2,Args1,Args2,LP).

checkhorpoeqnccMul(_,_,[],[]).
checkhorpoeqnccMul(N,Args1,[A2|Args2],[p(N1,N,equal)|LP]):-
	nthdelete1(N1,A1,Args1,Args3),
	eqlamTerm(A1,A2),!,
	N2 is N+1,
	checkhorpoeqnccMul(N2,Args3,Args2,LP).
checkhorpoeqnccMul(N,Args1,[A2|Args2],[p(N1,N,P)|LP]):-
	nth(N1,Args1,A1),
	checkhorponcc(A1,A2,P),!,
	N2 is N+1,
	checkhorpoeqnccMul(N2,Args1,Args2,LP).

checkhorponcc1FuncList(_,_,_,[],[]). 
checkhorponcc1FuncList(C,Args1,T1,[A2:T2|Args2],[p(0,P)|LP]):-
	geType(T1,T2),%!,
	checkhorponcc1Func(C,Args1,T1,A2,T2,P),
	checkhorponcc1FuncList(C,Args1,T1,Args2,LP).
checkhorponcc1FuncList(C,Args1,T1,[A2:T2|Args2],[p(N,equal)|LP]):-
	nth(N,Args1,A1),
        eqlamTerm(A1,A2:T2),
        checkhorponcc1FuncList(C,Args1,T1,Args2,LP),!.
checkhorponcc1FuncList(C,Args1,T1,[A2:T2|Args2],[p(N,P)|LP]):-
	nth(N,Args1,A1),
        checkhorponcc(A1,A2:T2,P),
        checkhorponcc1FuncList(C,Args1,T1,Args2,LP),!.

checkhorponcc1App(Args,_,Right,T2,p(5,N,equal)):- %Case 5 SubtApp
	nth(N,Args,A),
	eqlamTerm(A,Right:T2),!.
checkhorponcc1App(Args,T1,Right,T2,p(9,pf(LN),LP)):- %Case 9 MulApp
	nonvar(Right),
	Right=..[@|Args2],
	makeapp2(Args,T1,@(A1,A2)),
	getpartflat(Args2,T2,PF),
	length(PF,LN),
	checkhorponccMul(1,[A1,A2],PF,LP),!.
checkhorponcc1App(Args,_,Right,T2,p(5,N,P)):- %Case 5 SubtApp
	nth(N,Args,A),
	checkhorponcc(A,Right:T2,P),!.
checkhorponcc1App([A1,A2],T1,Right,T2,p(11,P)):- %Case 11 beta
	nonvar(A1),
	A1=lam(X,U),!,
	replace(X,A2,U,W),
	checkhorponcc(W:T1,Right:T2,P),!.
checkhorponcc1App([A1:(_->T),A2:_|Args],T1,Right,T2,p(11,P)):- %Case 11 beta
	nonvar(A1),
	A1=lam(X,U),!,
	replace(X,A2,U,W),
	checkhorponcc1App([W:T|Args],T1,Right,T2,P),!.
checkhorponcc1App(Args,T1,Right,T2,p(13,p(9,pf(2),LP))):- %Case 13 interm Beta
        nonvar(Right),
        Right=..[F|Args1],
        g_read(funcdecs,LFD),
        membereq2(F,LFD),
        makeapp2(Args,T1,@(A1,A2)),
        append(Args11,Ult:Tu,Args1),
        append(Args11,X:Tu,AArgs1),
        ARight=..[F|AArgs1],
        checkhorponccMul(1,[A1,A2],[lam(X,ARight:T2):(Tu->T2),Ult:Tu],LP),!.
        
checkhorponcc1Lam(lam(X,U),_,Right,T2,p(6,equal)):-  %Case  6 SubtLam
	replace(X,_,U,V),
	eqlamTerm(V,Right:T2),!.
checkhorponcc1Lam(lam(X,U),_,Right,T2,p(6,P)):-  %Case  6 SubtLam
	replace(X,_,U,V),
	checkhorponcc(V,Right:T2,P).
checkhorponcc1Lam(lam(X,U),(TI1->_),Right,T2,p(10,P)):-  %Case 10 MonLam
	nonvar(Right),
	nonvar(T2),
	Right=lam(Y,V),
	T2=(TI2->_),
	eqType(TI1,TI2),!,
	replace(Y,X,V,W),
	checkhorponcc(U,W,P).
checkhorponcc1Lam(lam(X,U),T1,Right,T2,p(12,P)):-  %Case 12 Eta
	nonvar(U),
	U=..[@|Args],
	append(Args,Args1,[Last]),
	Last==U,
	notinlist(X,Args1),!,
	makeapp(Args1,A),
	checkhorponcc1(A,T1,Right,T2,P),!.
checkhorponcc1Lam(lam(X,U),(TI1->_),Right,T2,p(14,P)):-  %Case 14 interm Eta
	nonvar(T2),
	T2=(TI2->TO2),!,
	eqType(TI1,TI2),!,
	replace(X,NV,U,V),
	addapp(Right,T2,NV:TI2,A),
	checkhorponcc(V,A:TO2,P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



eqlamTerm(X:T1,Y:T2):-
	(var(X);var(Y)),!,X==Y, eqType(T1,T2),!.
eqlamTerm(lam(X,U):T1,lam(Y,V):T2):-!,
	eqType(T1,T2),
	replace(Y,X,V,W),
	eqlamTerm(U,W).
eqlamTerm(A:T1,B:T2):-
	A=..[F|Args1],
	B=..[F|Args2],!,
	eqType(T1,T2),
	eqlamTermList(Args1,Args2).

eqlamTermList([],[]).
eqlamTermList([A1|Args1],[A2|Args2]):-
	eqlamTerm(A1,A2),
	eqlamTermList(Args1,Args2).

gtPrecF(C,C1):-
    g_read(funcprec,LP),
    gtPrec(C,C1,LP).

eqPrecF(C,C1):-
    g_read(funcprec,LP),
    eqPrec(C,C1,LP).

gtPrecT(C,C1):-
    g_read(typeprec,LP),
    gtPrec(C,C1,LP).

eqPrecT(C,C1):-
    g_read(typeprec,LP),
    eqPrec(C,C1,LP).

gtPrec(C,C1,LP):-
    nonvar(C),nonvar(C1),
    member([C,_,LG],LP),!,
    member(C1,LG).

eqPrec(C,C1,_):-
    (var(C);var(C1)),!,
    C==C1,!.
eqPrec(C,C1,_):-
    C==C1,!.
eqPrec(C,C1,LP):-
    member([C,LE,_],LP),!,
    member(C1,LE).

flatApp(A,Args1):-
    A=..[@,A1|Args],!,
    flatApp(A1,LA1),
    append(LA1,Args,Args1).
flatApp(A,[A]).

geType(X,Y):-
    X==Y.
geType(X,_):-
    var(X),!,fail.
geType(X,Y):-
    X=..[->,I1,O1],!,
    (geType(O1,Y);
     (nonvar(Y),Y=..[->,I2,O2],eqType(I1,I2),geType(O1,O2))).
geType(X,Y):-
    X=..[C|Args],
    g_read(typecons,LTC),
    membereqsol([C,_],LTC),!,
    geTypeCons(C,Args,Y).


geTypeCons(C,Args,Y):-
    nonvar(Y),
    Y=..[C1|Args1],
    g_read(typecons,LTC),
    membereqsol([C1,_],LTC),!,
    (gtPrecT(C,C1) -> gtTypeConstlist(C,Args,Args1) ;
     (eqPrecT(C,C1) -> geTypeStat(C,Args,Args1) ;
      (member(T,Args),geType(T,Y)))).
geTypeCons(_,Args,Y):-
	member(T,Args),
	geType(T,Y).

gtTypeConstlist(_,_,[]).
gtTypeConstlist(C,Args,[A1|Args1]):-
	geTypeCons(C,Args,A1),
	gtTypeConstlist(C,Args,Args1).

gtType(X,Y):-
    X==Y,!,fail.
gtType(X,_):-
    var(X),!,fail.
gtType(X,Y):-
    X=..[->,I1,O1],!,
    (gtType(O1,Y);
     (nonvar(Y),Y=..[->,I2,O2],eqType(I1,I2),gtType(O1,O2))).
gtType(X,Y):-
    X=..[C|Args],
    g_read(typecons,LTC),
    membereqsol([C,_],LTC),!,
    gtTypeCons(C,Args,Y).


gtTypeCons(C,Args,Y):-
    nonvar(Y),
    Y=..[C1|Args1],
    g_read(typecons,LTC),
    membereqsol([C1,_],LTC),!,
    (gtPrecT(C,C1) -> gtTypeConstlist(C,Args,Args1) ;
     (eqPrecT(C,C1) -> gtTypeStat(C,Args,Args1) ;
      (member(T,Args),geType(T,Y)))).
gtTypeCons(_,Args,Y):-
	member(T,Args),
	geType(T,Y).

geTypeStat(C,Args,Args1):-
    eqTypeStat(C,Args,Args1),!.
geTypeStat(C,Args,Args1):-
    gtTypeStat(C,Args,Args1),!.

eqType(I1,I2):-(var(I1);var(I2)),!,I1=I2.
eqType(I1,I2):-
	I1=..[->|Args1],
	I2=..[->|Args2],
	eqTypeList(Args1,Args2).
eqType(I1,I2):-
	I1=..[C1|Args1],
	I2=..[C2|Args2],
	eqPrecT(C1,C2),
	eqTypeStat(C1,Args1,Args2).

eqTypeStat(C,Args1,Args2):-
    g_read(typelex,[L1,L2]),
    (member(C,L1);member(C,L2)),
    eqTypeList(Args1,Args2).
eqTypeStat(_,Args1,Args2):-
    eqTypeMul(Args1,Args2).

eqTypeList([],[]). 
eqTypeList([T1|Args1],[T2|Args2]):-
    eqType(T1,T2),
    eqTypeList(Args1,Args2). 


eqTypeMul([],[]). 
eqTypeMul([T1|Args1],Args2):-
    delete1(T2,Args2,Args3),
    eqType(T1,T2),!,
    eqTypeMul(Args1,Args3). 


gtTypeStat(C,Args1,Args2):-
    g_read(typelex,[L1,L2]),
    (member(C,L1);member(C,L2)),
    gtTypeLex(Args1,Args2,Args3),
    gtTypeConstlist(C,Args1,Args3).
gtTypeStat(_,Args1,Args2):-
    gtTypeMul(Args1,Args2).

gtTypeLex([T1|_],[T2|Args2],Args2):-
    gtType(T1,T2). 
gtTypeLex([T1|Args1],[T2|Args2],Args3):-
    eqType(T1,T2),
    gtTypeLex(Args1,Args2,Args3). 


gtTypeMul([_|_],[]). 
gtTypeMul(Args1,[T2|Args2]):-
    delete1(T1,Args1,Args3),
    eqType(T1,T2),!,
    gtTypeMul(Args3,Args2). 
gtTypeMul(Args1,[T2|Args2]):-
    member(T1,Args1),
    gtType(T1,T2),!,
    geTypeMul(Args1,Args2). 

geTypeMul([],[]). 
geTypeMul(Args1,[T2|Args2]):-
    delete1(T1,Args1,Args3),
    eqType(T1,T2),!,
    geTypeMul(Args3,Args2). 
geTypeMul(Args1,[T2|Args2]):-
    member(T1,Args1),
    gtType(T1,T2),!,
    geTypeMul(Args1,Args2). 



write_lamterm(T):-flatlamapp(T,FT),write(FT),!.

flatlamapp(T,T):-var(T),!.
flatlamapp('$VAR'(N),'$VAR'(N)):-!.
flatlamapp(T,FT):-
	T=..[@|_],!,
	flatApp(T,Args),
	flatlamapplist(Args,Args1),
	FT=..[@|Args1].
flatlamapp(T,FT):-
	T=..[lam|_],!,
	flatLam(T,Vars,Body),
	flatlamapp(Body,Body1),
	append(Vars,[Body1],Args),
	FT=..[lam|Args].
flatlamapp(T,FT):-
	T=..[F|Args],
	flatlamapplist(Args,Args1),
	FT=..[F|Args1].

flatlamapplist([],[]).
flatlamapplist([A|Args],[A1|Args1]):-
	flatlamapp(A,A1),
	flatlamapplist(Args,Args1).


flatLam(lam(X,U):_,[X|Vars],Body):-!,
	flatLam(U,Vars,Body).
flatLam(lam(X,U),[X|Vars],Body):-!,
	flatLam(U,Vars,Body).
flatLam(Body,[],Body).


delete1(X,[X|L],L).
delete1(X,[Y|L1],[Y|L2]):-
    delete1(X,L1,L2).

nthdelete1(N,X,[E|L1],[[]|L2]):-
	E==[],
	nthdelete1(N1,X,L1,L2),
	N is N1 + 1.
nthdelete1(1,X,[X|L],[[]|L]).
nthdelete1(N,X,[Y|L1],[Y|L2]):-
    nthdelete1(N1,X,L1,L2),
    N is N1 + 1.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


writeblanks(0):-!.
writeblanks(N):-
    write(' '),
    N1 is N-1,
    writeblanks(N1),!.

writeNth(1):- write('1st'),!.
writeNth(2):- write('2nd'),!.
writeNth(3):- write('3rd'),!.
writeNth(N):- N<21,write(N),write('th'),!.
writeNth(N):- 1 is N mod 10, write(N),write('st'),!.
writeNth(N):- 2 is N mod 10, write(N),write('nd'),!.
writeNth(N):- 3 is N mod 10, write(N),write('rd'),!.
writeNth(N):- write(N),write('th'),!.

prettyprint(N,p(1,cc,p(cc,equal,P))):-!, %Case 1 CC
	writeblanks(N),
	write('By case 1, the rhs is included in Computable Closure:'),nl,
	N1 is N+3,
	prettyprintCC(N1,P),!.
prettyprint(N,p(1,NA,equal)):-!, %Case 1 Subterm
	writeblanks(N),
	write('By case 1, the '), writeNth(NA), 
	write(' argument is equal to the rhs.'),
	nl,!.
prettyprint(N,p(1,NA,P)):-	%Case 1 Subterm
	writeblanks(N),
	write('By case 1, the '),
	writeNth(NA), write(' argument is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(2,[])):-	%Case 2 Prec 
	writeblanks(N),
	write('By case 2.'),nl,!.
prettyprint(N,p(2,LP)):-	%Case 2 Prec 
	writeblanks(N),
	write('By case 2:'),nl,
	N1 is N+3,
	prettyprintList(N1,1,LP),!.
prettyprint(N,p(3,LP)):-	%Case 3 Mul %Case 4 Lex
	writeblanks(N),
	write('By case 3:'),nl,
	N1 is N+3,
	prettyprintMul(N1,LP),!.
prettyprint(N,p(4-lr,LP)):-	%Case 3 Mul %Case 4 Lex
	writeblanks(N),
	write('By case 4 (left-to-right):'),nl,
	N1 is N+3,
	prettyprintLexLR(N1,LP),!.
prettyprint(N,p(4-rl,LP)):-	%Case 3 Mul %Case 4 Lex
	writeblanks(N),
	write('By case 4 (right-to-left):'),nl,
	N1 is N+3,
	prettyprintLexRL(N1,LP),!.
prettyprint(N,p(5,NA,equal)):-!, %Case 5 SubtApp
	writeblanks(N),
	write('By case 5, the '), writeNth(NA), 
	write(' argument is equal to the rhs.'),
	nl,!.
prettyprint(N,p(5,NA,P)):-	%Case 5 SubtApp
	writeblanks(N),
	write('By case 5, the '),
	writeNth(NA), write(' argument is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(6,equal)):-  %Case  6 SubtLam
	writeblanks(N),
	write('By case 6, the body is equal to the rhs.'),
	nl,!.
prettyprint(N,p(6,P)):-  %Case  6 SubtLam
	writeblanks(N),
	write('By case 6, the body of the lhs is greter than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(7,LP)):-	%Case 7 FuncApp
	writeblanks(N),
	write('By case 7:'),nl,
	N1 is N+3,
	prettyprintList(N1,1,LP),!.
prettyprint(N,p(8,P)):-		%Case 8 FuncLam
	writeblanks(N),
	write('By case 8:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(9,pf(LN),LP)):- %Case 9 MulApp
	writeblanks(N),
	write('By case 9, with a partial flattening of length '), write(LN), write(':'),nl,
	N1 is N+3,
	prettyprintMul(N1,LP),!.
prettyprint(N,p(10,P)):-  %Case 10 MonLam
	writeblanks(N),
	write('By case 10, the body of the lhs is greter than the body of the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(11,P)):- %Case 11 beta
	writeblanks(N),
	write('By case 11, the beta-reduction of the lhs is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(12,P)):-  %Case 12 Eta
	writeblanks(N),
	write('By case 12, the eta-reduction of the lhs is greater than the rhs:'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.
prettyprint(N,p(13,p(9,pf(2),LP))):- %Case 13 interm Beta
	writeblanks(N),
	write('We build a term that beta-reduces (by case 11) to the rhs by abstracting'),nl,
	writeblanks(N),
        write('this rhs over its last argument and show that it is smaller than the lhs'),nl,
	writeblanks(N),
        write('by case 9 with a partial flattening of length 2:'),nl,
	N1 is N+3,
	prettyprintMul(N1,LP),!.
prettyprint(N,p(14,P)):-  %Case 14 interm Eta
	writeblanks(N),
	write('The term that eta-reduces (in one step) to the rhs (by case 12) and'),nl,
	writeblanks(N),
        write('is smaller than the lhs by case 10'),nl,
	N1 is N+3,
	prettyprint(N1,P),!.

prettyprintLexLR(NI,[p(NA,P)]):-!,
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P).
prettyprintLexLR(NI,[p(NA,P),restargs,NR,LP]):-!,
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	writeblanks(NI),
	write('For the rest of the arguments we have:'),nl,
	prettyprintList(NI1,NR,LP).
prettyprintLexLR(NI,[p(NA,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is equal to the one of the rhs:'),nl,
	prettyprintLexLR(NI,LP),!.

prettyprintLexRL(NI,[p(NA,P)]):-!,
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P).
prettyprintLexRL(NI,[p(NA,P),restargs,_,LP]):-!,
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	writeblanks(NI),
	write('For the rest of the arguments we have:'),nl,
        reverse(LP,LP1),
	prettyprintList(NI1,1,LP1).
prettyprintLexRL(NI,[p(NA,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is equal to the one of the rhs:'),nl,
	prettyprintLexRL(NI,LP),!.

writeLast(0):-write('last'),!.
writeLast(N):-write('last-'),write(N),!.

prettyprintMul(_,[]).
prettyprintMul(NI,[p(N1,N,equal)|LP]):-!,
	writeblanks(NI),
	write('- The '),writeNth(N1), 
	write(' element of the lhs is equal to the '),
	writeNth(N), write(' element of the rhs.'), nl,
	prettyprintMul(NI,LP).
prettyprintMul(NI,[p(N1,N,P)|LP]):-
	writeblanks(NI),
	write('- The '),writeNth(N1), 
	write(' element of the lhs is greater than '),
	writeNth(N), write(' element of the rhs.'), nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	prettyprintMul(NI,LP).

prettyprintList(_,_,[]). 
prettyprintList(NI,N,[p(cc,p(cc,equal,P))|LP]):-!,
	writeblanks(NI),
	write('- The '),writeNth(N),
	write(' argument of the rhs is in included in Computable Closure:'),nl,
	NI1 is NI+3,
	prettyprintCC(NI1,P),
	N1 is N+1,
	prettyprintList(NI,N1,LP).
prettyprintList(NI,N,[p(0,P)|LP]):-
	writeblanks(NI),
	write('- The lhs is greater than the '),writeNth(N),
	write(' argument of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	N1 is N+1,
	prettyprintList(NI,N1,LP).
prettyprintList(NI,N,[p(Nr,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeNth(Nr),
	write(' argument of the lhs is equal to the '),
	writeNth(N),
	write(' argument of the rhs.'),nl,
	N1 is N+1,
	prettyprintList(NI,N1,LP).
prettyprintList(NI,N,[p(Nr,P)|LP]):-
	writeblanks(NI),
	write('- The '), writeNth(Nr),
	write(' argument of the lhs is greater than the '),
	writeNth(N),
	write(' argument of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	N1 is N+1,
	prettyprintList(NI,N1,LP).

prettyprintCC(NI,pcc(bvar,base)):-
        writeblanks(NI),
	write('By base case, since it is in the set of bound vars V.'),nl,!.
prettyprintCC(NI,pcc(N,base)):-
	writeblanks(NI),
	write('By base case, since it is the '),writeNth(N),
        write(' argument.'),nl,!.
prettyprintCC(NI,pcc(abs,_,P)):- %abstraction
	writeblanks(NI),
	write('By abstraction case, adding (after renaming) the bound variable to V and'),nl,
	writeblanks(NI),
        write('the body is inside the Computable Closure:'),nl,
	NI1 is NI+3,
	prettyprintCC(NI1,P),!.
prettyprintCC(NI,pcc(app,LP)):- %application
	writeblanks(NI),
	write('By application case, showing that the arguments are in the Computable Closure:'),nl,
	NI1 is NI+3,
        prettyprintCClist(NI1,1,LP),!.
prettyprintCC(NI,pcc(prec,[])):- %precedence
	writeblanks(NI),
	write('By precedence case, since it has no arguments'),nl,!.
prettyprintCC(NI,pcc(prec,LP)):- %precedence
	writeblanks(NI),
	write('By precedence case, showing that the arguments are in the Computable Closure:'),nl,
	NI1 is NI+3,
        prettyprintCClist(NI1,1,LP),!.
prettyprintCC(NI,pcc(reccall,mul,OP,LP)):- %recursive-call
	writeblanks(NI),
	write('By recursive-call case, showing that the arguments decrease with'),nl,
	writeblanks(NI),
        write('the multiset extension of horpo union subterm of decreasing type:'),nl,
	NI1 is NI+3,
        prettyprintsubtMul(NI1,OP),
	writeblanks(NI),
	write('and showing that the arguments are in the Computable Closure:'),nl,
        prettyprintCClist(NI1,1,LP),!.
prettyprintCC(NI,pcc(reccall,lex-lr,OP,LP)):- %recursive-call
	writeblanks(NI),
	write('By recursive-call case, showing that the arguments decrease with'),nl,
	writeblanks(NI),
        write('the lex (left-to-right) extension of horpo union subterm of decreasing type:'),nl,
	NI1 is NI+3,
        prettyprintsubtLexLR(NI1,OP),
	writeblanks(NI),
	write('and showing that the arguments are in the Computable Closure:'),nl,
        prettyprintCClist(NI1,1,LP),!.
prettyprintCC(NI,pcc(reccall,lex-rl,OP,LP)):- %recursive-call
	writeblanks(NI),
	write('By recursive-call case, showing that the arguments decrease with'),nl,
	writeblanks(NI),
        write('the lex (right-to-left) extension of horpo union subterm of decreasing type:'),nl,
	NI1 is NI+3,
        prettyprintsubtLexRL(NI1,OP),
	writeblanks(NI),
	write('and showing that the arguments are in the Computable Closure:'),nl,
        prettyprintCClist(NI1,1,LP),!.
prettyprintCC(NI,pcc(red-base,N,OP)):- %reduction
	writeblanks(NI),
	write('By reduction case, showing that is smaller than the '), writeNth(N), 
        write(' argument:'),nl,
	NI1 is NI+3,
        prettyprint(NI1,OP),!.

prettyprintCClist(_,_,[]).
prettyprintCClist(NI,N,[P|LP]):-
	writeblanks(NI),
        write('- The '), writeNth(N),
	write(' argument is in the Computable Closure: '),nl,
	NI1 is NI+3,
        prettyprintCC(NI1,P),
        N1 is N + 1,
        prettyprintCClist(NI,N1,LP),!.

prettyprintsubtLexLR(NI,[p(NA,P)]):-!,
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P).
prettyprintsubtLexLR(NI,[p(NA,substgT)]):-!,
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the rhs is a subterm of decreasing type of the one of the lhs.'),
        nl,!.
prettyprintsubtLexLR(NI,[p(NA,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeNth(NA),
	write(' argument of the lhs is equal to the one of the rhs:'),nl,
	prettyprintsubtLexLR(NI,LP),!.

prettyprintsubtLexRL(NI,[p(NA,P)]):-!,
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is greater than the one of the rhs:'),nl,
	NI1 is NI+3,
	prettyprint(NI1,P).
prettyprintsubtLexRL(NI,[p(NA,substgT)]):-!,
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the rhs is a subterm of decreasing type of the one of the lhs.'),nl,!.
prettyprintsubtLexRL(NI,[p(NA,equal)|LP]):-
	writeblanks(NI),
	write('- The '), writeLast(NA),
	write(' argument of the lhs is equal to the one of the rhs:'),nl,
	prettyprintsubtLexRL(NI,LP),!.

prettyprintsubtMul(_,[]).
prettyprintsubtMul(NI,[p(N1,N,equal)|LP]):-!,
	writeblanks(NI),
	write('- The '),writeNth(N1), 
	write(' element of the lhs is equal to the '),
	writeNth(N), write(' element of the rhs.'), nl,
	prettyprintsubtMul(NI,LP).
prettyprintsubtMul(NI,[p(N1,N,substgT)|LP]):-
	writeblanks(NI),
	write('- The '),writeNth(N), 
	write(' element of the rhs is subterm of decresing type of the '),nl,
	writeblanks(NI),
	writeNth(N1), write(' element of the lhs.'), nl,
	prettyprintsubtMul(NI,LP).
prettyprintsubtMul(NI,[p(N1,N,P)|LP]):-
	writeblanks(NI),
	write('- The '),writeNth(N1), 
	write(' element of the lhs is greater than '),
	writeNth(N), write(' element of the rhs.'), nl,
	NI1 is NI+3,
	prettyprint(NI1,P),
	prettyprintsubtMul(NI,LP).
